#include <iostream>
#include <math.h>

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
#include <windows.h>
#endif

#include <GL/gl.h>

#include "Tools.h"
#include "CA2D.h"

CA2D::CA2D(float xPosTopLeft, float yPosTopLeft, float xSize, float ySize, 
	float cornerRadius, float visualUpdateStepSize, float colorChangeTC, 
	float stateUpdateStepSize)
{
	mXPosTopLeft = xPosTopLeft;
	mYPosTopLeft = yPosTopLeft;
	mXSize = xSize;
	mYSize = ySize;
	mRadius = cornerRadius;

	// Use the color change time constant to compute a multiplicative scalar.
	mColorChangeConstant = ::expf(-visualUpdateStepSize / colorChangeTC);

	for (size_t i = 0; i < 8; ++i)
	{
		mNeighbors[i] = NULL;
	}

	mActive = false;
	mNextStepActive = false;
	mActiveTimer = 0;
	mColor = 0;
	mStateUpdateStepSize = stateUpdateStepSize;

	assert(xSize >= 2 * cornerRadius);
	assert(ySize >= 2 * cornerRadius);

/*
	// Create a GL display list for the visual shape.
	// TODO: if things run too slow, we could compile a single 
	// display list and share it among all CAs.
	mDisplayList = glGenLists(1);
	glNewList((GLuint)mDisplayList, GL_COMPILE);
	glBegin(GL_QUADS);
		// TODO: rounded corners
		glVertex2f(0, 0);
		glVertex2f(0, -1);
		glVertex2f(1, -1);
		glVertex2f(1, 0);
	glEnd();
	glEndList();
*/
}

CA2D::~CA2D()
{
}

bool CA2D::isActive()const
{
	return mActive;
}

void CA2D::setActivity(bool active)
{
	mActive = active;
	mNextStepActive = active;
	mActiveTimer = 0.5f;
}

void CA2D::setNeighbor(size_t neighborIndex, CA2D* neighbor)
{
	assert(neighborIndex < 8);
	mNeighbors[neighborIndex] = neighbor;
}

void CA2D::updateStatePhase1()
{
	size_t numActiveNeighbors = 0;
	for (size_t i = 0; i < 8; ++i)
	{
		if (mNeighbors[i] && mNeighbors[i]->isActive())
		{
			++numActiveNeighbors;
		}
	}

	// Compute the new state based on the neighbors' states.  Use the 
	// "Conway's game of life" rules.
	if (mActive)
	{
		if (numActiveNeighbors <= 1)
		{
			// Each active cell with one or no active neighbors dies, as if by 
			// loneliness.
			mNextStepActive = false;
		}
		else if (numActiveNeighbors >= 4)
		{
			// Each active cell with four or more active neighbors dies, as 
			// if by overpopulation.
			mNextStepActive = false;
		}

		// Each active cell with two or three active neighbors survives.
	}
	else
	{
		if (3 == numActiveNeighbors)
		{
			// Each inactive cell with three active neighbors becomes 
			// populated.
			mNextStepActive = true;
		}
	}

	if (mActiveTimer > 0)
	{
		mNextStepActive = true;
		mActiveTimer -= mStateUpdateStepSize;
	}
}

void CA2D::updateStatePhase2()
{
	mActive = mNextStepActive;
}

void CA2D::updateColor()
{
	// Update the color using a leaky integrator system defined by a time 
	// constant.
	float targetColor = 0;
	if (mActive)
	{
		targetColor = 1;
	}
	mColor = mColorChangeConstant * mColor + (1 - mColorChangeConstant) * 
		targetColor;
}

void CA2D::draw()
{
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(mXPosTopLeft, mYPosTopLeft, 0);
	glScalef(mXSize, mYSize, 0);
	//glColor3f(mColor, mColor, mColor);
	glColor3f(0.2 * mColor, 0.3f + 0.6f * mColor, 0.5 + 0.3f * mColor);
	//glCallList((GLuint)mDisplayList);

	float radius = mRadius / mXSize;
	size_t numPoints = 6;
	assert(numPoints >= 2);
	glBegin(GL_POLYGON);
		// Top-left corner.
		float cornerCenterX = radius;
		float cornerCenterY = -radius;
		float angleIncrement = 90 / ((float)numPoints - 1);
		for (float angle = 90; angle <= 180; angle += angleIncrement)
		{
			float vecX = radius;
			float vecY = 0;
			rotateVec2D(vecX, vecY, angle);
			glVertex3f(cornerCenterX + vecX, cornerCenterY + vecY, 0);
		}

		// Bottom-left corner.
		cornerCenterX = radius;
		cornerCenterY = -1 + radius;
		for (float angle = 180; angle <= 270; angle += angleIncrement)
		{
			float vecX = radius;
			float vecY = 0;
			rotateVec2D(vecX, vecY, angle);
			glVertex3f(cornerCenterX + vecX, cornerCenterY + vecY, 0);
		}

		// Bottom-right corner.
		cornerCenterX = 1 - radius;
		cornerCenterY = -1 + radius;
		for (float angle = 270; angle <= 360; angle += angleIncrement)
		{
			float vecX = radius;
			float vecY = 0;
			rotateVec2D(vecX, vecY, angle);
			glVertex3f(cornerCenterX + vecX, cornerCenterY + vecY, 0);
		}

		// Top-right corner.
		cornerCenterX = 1 - radius;
		cornerCenterY = -radius;
		for (float angle = 360; angle <= 450; angle += angleIncrement)
		{
			float vecX = radius;
			float vecY = 0;
			rotateVec2D(vecX, vecY, angle);
			glVertex3f(cornerCenterX + vecX, cornerCenterY + vecY, 0);
		}
	glEnd();

/*
	glBegin(GL_QUADS);
		// TODO: rounded corners
		glVertex2f(0, 0);
		glVertex2f(0, -1);
		glVertex2f(1, -1);
		glVertex2f(1, 0);
	glEnd();
*/

	glPopMatrix();
}
