// Cellular Automata Touch
// Tyler Streeter
// Fall 2007

#include <time.h>

#include <SDL/SDL.h>

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include "CA2D.h"
#include "Tools.h"
#include "InputEventListener.h"

InputEventListener* gInputSystem = NULL;

std::vector<CA2D*> gCAs;
float gCAStateUpdatePeriod = 0.2f; // 5 Hz
float gCAVisualUpdatePeriod = 0.0333f; // 30 Hz
float gCAStateUpdateTimer = 0;
float gCAVisualUpdateTimer = 0;
size_t gNumCellsX = 40;
size_t gNumCellsY = 0;
size_t gWindowWidth = 1024;
size_t gWindowHeight = 768;
bool gQuit = false;
bool gPaused = false;
bool gSlowMotion = false;
bool gKeyStates[323];
float gViewTrans[2] = {0.0f, 0.0f};
float gViewTransSpeed = 0.2f;
float gViewRot[2] = {0.0f, 0.0f};
float gViewRotSpeed = 0.2f;
float gViewZoom = -50.0f;
float gViewZoomSpeed = 0.2f;

void mainLoop();
void setupWindow(size_t windowWidth, size_t windowHeight);
void destroyWindow();
bool processSDLInput();
void processTouchInput();

int main(int argc, char **argv)
{
	if(argc != 4)
	{
		std::cout << "Usage: executable [url/ip] [publish port] [event request port]\n";
	}

	std::string host(argv[1]);
	std::cout << "Starting application, host: " << host 
		<< ",  publish port: " << atoi(argv[2]) << ", request port: " 
		<< atoi(argv[3]) << std::endl;
	gInputSystem = new InputEventListener(host, atoi(argv[2]), atoi(argv[3]), 
		gWindowWidth, gWindowHeight);

	srand((unsigned int)time(NULL));

	for (size_t i = 0; i < 323; ++i)
	{
		gKeyStates[i] = false;
	}

	// Compute the number of cells in x and y, the spacing between cells, 
	// the cell dimensions, etc.  Note: we assume the viewport size is 2x2.
	gNumCellsY = (size_t)(((float)gWindowHeight / (float)gWindowWidth) * gNumCellsX);
	const float xSpacing = 2 / (float)gNumCellsX;
	const float ySpacing = 2 / (float)gNumCellsY;
	const float xSize = 0.9f * xSpacing;
	const float ySize = 0.9f * ySpacing;
	const float cornerRadius = 0.1f * xSize;
	const float colorChangeTC = 0.1f;

	// Create CAs.
	float xPosTopLeft = -1 + 0.5f * 0.1f * xSpacing;
	float yPosTopLeft = 1 - 0.5f * 0.1f * ySpacing;
	for (size_t j = 0; j < gNumCellsY; ++j)
	{
		for (size_t i = 0; i < gNumCellsX; ++i)
		{
			gCAs.push_back(new CA2D(xPosTopLeft, yPosTopLeft, xSize, ySize, 
				cornerRadius, gCAVisualUpdatePeriod, colorChangeTC, 
				gCAStateUpdatePeriod));

			xPosTopLeft += xSpacing;
		}

		xPosTopLeft = -1 + 0.5f * 0.1f * xSpacing;
		yPosTopLeft -= ySpacing;
	}

	std::cout << "Created " << gCAs.size() << " cells (" 
		<< gNumCellsX << " x " << gNumCellsY << ")" << std::endl;

	// Setup neighbor connections.
	for (size_t j = 0; j < gNumCellsY; ++j)
	{
		for (size_t i = 0; i < gNumCellsX; ++i)
		{
			size_t neighborIndex = 0;

			for (int ny = (int)j - 1; ny <= (int)j + 1; ++ny)
			{
				for (int nx = (int)i - 1; nx <= (int)i + 1; ++nx)
				{
					if (!(nx == i && ny == j))
					{
						if (nx >= 0 && nx < (int)gNumCellsX && 
							ny >= 0 && ny < (int)gNumCellsY)
						{
							gCAs[j * gNumCellsX + i]->setNeighbor(
								neighborIndex, gCAs[ny * gNumCellsX + nx]);
						}

						++neighborIndex;
					}
				}
			}
		}
	}

	//for (size_t i = 0; i < gCAs.size(); ++i)
	//{
	//	size_t numNeighbors = 0;
	//	for (size_t n = 0; n < 26; ++n)
	//	{
	//		if (gCAs[i]->mNeighbors[n])
	//		{
	//			++numNeighbors;
	//		}
	//	}

	//	std::cout << "CA " << i << " has " << numNeighbors << " neighbors" << std::endl;
	//}

	setupWindow(gWindowWidth, gWindowHeight);
	mainLoop();
	destroyWindow();

	// Destroy the CAs.
	while (!gCAs.empty())
	{
		delete gCAs.back();
		gCAs.pop_back();
	}

	if (gInputSystem)
	{
		delete gInputSystem;
	}

	return 0;
}

void mainLoop()
{
	unsigned int prevTime_ms = SDL_GetTicks();
	while (!gQuit)
	{
		// Compute the dt since the previous frame.
		unsigned int currentTime_ms = SDL_GetTicks();
		double dt_sec = 0.001 * (currentTime_ms - prevTime_ms);
		prevTime_ms = currentTime_ms;

		if (gSlowMotion)
		{
			dt_sec *= 0.125;
		}

		gQuit = processSDLInput();
		processTouchInput();

		if (!gPaused)
		{
			gCAStateUpdateTimer += dt_sec;
			if (gCAStateUpdateTimer >= gCAStateUpdatePeriod)
			{
				while(gCAStateUpdateTimer > gCAStateUpdatePeriod)
				{
					gCAStateUpdateTimer -= gCAStateUpdatePeriod;
				}

				// We perform a 2-stage update on the cells because the first 
				// update depends on all the previous states.
				size_t numCells = gCAs.size();
				for (size_t i = 0; i < numCells; ++i)
				{
					gCAs[i]->updateStatePhase1();
				}
				for (size_t i = 0; i < numCells; ++i)
				{
					gCAs[i]->updateStatePhase2();
				}
			}

			gCAVisualUpdateTimer += dt_sec;
			if (gCAVisualUpdateTimer >= gCAVisualUpdatePeriod)
			{
				while(gCAVisualUpdateTimer > gCAVisualUpdatePeriod)
				{
					gCAVisualUpdateTimer -= gCAVisualUpdatePeriod;
				}

				size_t numCells = gCAs.size();
				for (size_t i = 0; i < numCells; ++i)
				{
					gCAs[i]->updateColor();
				}
			}
		}

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();

		// Orbit cam.
		//glTranslatef(gViewTrans[0], gViewTrans[1], gViewZoom);
		//glRotatef(gViewRot[0], 1.0, 0.0, 0.0);
		//glRotatef(gViewRot[1], 0.0, 1.0, 0.0);

		//glDisable(GL_LIGHTING);
		//glDisable(GL_DEPTH_TEST);
		for (size_t i = 0; i < gCAs.size(); ++i)
		{
			gCAs[i]->draw();
		}
		//glEnable(GL_DEPTH_TEST);
		//glEnable(GL_LIGHTING);

		// Do OpenGL swap buffers (via SDL).
		SDL_GL_SwapBuffers();
	}
}

void setupWindow(size_t windowWidth, size_t windowHeight)
{
	float aspectRatio = (float)windowWidth / (float)windowHeight;

	if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		std::cerr << "SDL initialization failed: " << SDL_GetError() 
			<< std::endl;
		SDL_Quit();
		return;
	}

	SDL_Surface *SDLSurface = SDL_SetVideoMode((int)windowWidth, (int)windowHeight, 
		0, SDL_OPENGL|SDL_RESIZABLE);

	// Enable double buffering.
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	if (NULL == SDLSurface)
	{
		fprintf(stderr, "Video mode initialization failed: %s\n", SDL_GetError());
		SDL_Quit();
		return;
	}

	// Set key repeat rate (initial delay and interval between repeats, 
	// both in milliseconds).
	SDL_EnableKeyRepeat(200, 200);

	// Setup OpenGL.

	glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3f(0.0f, 0.0f, 0.0f);

	glViewport(0, 0, (GLsizei)windowWidth, (GLsizei)windowHeight);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//gluPerspective(40.0f, aspectRatio, 1, 5000.0f);
	glOrtho(-1, 1, -1, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW);

	//glEnable(GL_DEPTH_TEST);

	// Don't draw the back of faces.
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	//glEnable(GL_NORMALIZE);

	// Don't light the back of faces.
	//glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 0);
	//glEnable(GL_LIGHTING);

	// Enable blending.
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

/*
	// Create a light source.
	GLfloat ambient[] = {0.5f, 0.5f, 0.3f, 1.0f};
	GLfloat diffuse[] = {0.9f, 0.9f, 0.9f, 1.0f};
	GLfloat specular[] = {1.0f, 1.0f, 1.0f, 1.0f};
	GLfloat light_position[] = {-50, 10, 50, 0.0f};
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
*/
}

void destroyWindow()
{
	if(SDL_WasInit(SDL_INIT_VIDEO))
	{
		SDL_Quit();
	}
}

void processTouchInput()
{
	while (gInputSystem->hasUnhandledTouchPoints())
	{
		float normalizedX = 0;
		float normalizedY = 0;
		gInputSystem->popTouchPoint(normalizedX, normalizedY);

		// Hack to ignore the final 0, 0 point which occurs after each touch.
		if (0 == normalizedX && 0 == normalizedY)
		{
			continue;
		}

		// Compute the 2D cell index.
		size_t cellIndexX = (size_t)(normalizedX * gNumCellsX);
		if (cellIndexX == gNumCellsX)
		{
			--cellIndexX;
		}
		size_t cellIndexY = (size_t)(normalizedY * gNumCellsY);
		if (cellIndexY == gNumCellsY)
		{
			--cellIndexY;
		}

		// Convert the 2D index to a 1D index.
		size_t cellIndex = cellIndexY * gNumCellsX + cellIndexX;

		std::cout << "Point " << normalizedX << ", " << normalizedY << ", activating cell " << cellIndex << '\n';

		// Make the touched cell active.
		gCAs[cellIndex]->setActivity(true);
	}
}

bool processSDLInput()
{
	bool quit = false;

	// Wandle events until the SDL event queue is empty.
	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		switch(event.type)
		{
			case SDL_KEYDOWN:
				switch(event.key.keysym.sym)
				{
					case SDLK_ESCAPE:
						quit = true;
						break;
					case SDLK_q:
						quit = true;
						break;
					case SDLK_r:
						std::cout << "Resetting activity...\n";
						for (size_t i = 0; i < gCAs.size(); ++i)
						{
							gCAs[i]->setActivity(false);
						}
						break;
					case SDLK_1:
						std::cout << "Randomizing activity...\n";
						for (size_t i = 0; i < gCAs.size(); ++i)
						{
							if (randomRealUniform(0, 1) > 0.8f)
							{
								gCAs[i]->setActivity(true);
							}
							else
							{
								gCAs[i]->setActivity(false);
							}
						}
						break;
					default:
						break;
				}
				break;
			/*
			case SDL_VIDEORESIZE:
			{
				setupWindow(event.resize.w, event.resize.h);
				break;
			}
			*/
			case SDL_QUIT:
			{
				quit = true;
				break;
			}
			default:
				break;
		}
	}

	// Apply toggleable states.
	//if (gKeyStates[SDLK_UP])
	//{
	//}

	return quit;
}
